#include <poppler.h>
#include "ckd-page.h"

G_DEFINE_TYPE (CkdPage, ckd_page, CLUTTER_TYPE_ACTOR);

#define CKD_PAGE_GET_PRIVATE(obj) (G_TYPE_INSTANCE_GET_PRIVATE ((obj), CKD_TYPE_PAGE, CkdPagePriv))

typedef struct _CkdPagePriv {
        PopplerPage *pdf_page;
        gfloat scale;
        ClutterActor *cairo_texture;
} CkdPagePriv;

enum {
        PROP_0,
        PROP_PDF_PAGE,
        PROP_SCALE
};

static void ckd_page_get_aspect (PopplerPage *page, gfloat *width, gfloat *height, gfloat *aspect);
static void ckd_page_get_parent_size (ClutterActor *actor, gfloat *width,
                                      gfloat *height, gfloat *aspect);

static gboolean
ckd_page_render (ClutterCairoTexture *actor, cairo_t *cr, gpointer user_data)
{
        CkdPage *self = CKD_PAGE (user_data);
        CkdPagePriv *priv = CKD_PAGE_GET_PRIVATE (self);

        clutter_cairo_texture_clear (actor);

        if (priv->pdf_page) {
                cairo_scale (cr, priv->scale, priv->scale);
                poppler_page_render (priv->pdf_page, cr);
        }
        return TRUE;
}

static void
ckd_page_paint (ClutterActor *actor)
{
        CkdPage *self = CKD_PAGE (actor);
        CkdPagePriv *priv = CKD_PAGE_GET_PRIVATE (self);
        clutter_actor_paint (priv->cairo_texture);
}

static void
ckd_page_set_property (GObject *obj, guint prop_id, const GValue *value, GParamSpec *pspec)
{
        CkdPage *self = CKD_PAGE (obj);
        CkdPagePriv *priv = CKD_PAGE_GET_PRIVATE (self);
        gdouble w, h;

        switch (prop_id) {
        case PROP_PDF_PAGE:
                priv->pdf_page = g_value_get_pointer (value);
                if (priv->cairo_texture) {
                        clutter_actor_destroy (priv->cairo_texture);
                }
                poppler_page_get_size (priv->pdf_page, &w, &h);
                priv->cairo_texture = clutter_cairo_texture_new (w, h);
                clutter_actor_set_parent (priv->cairo_texture, CLUTTER_ACTOR (self));
                clutter_cairo_texture_set_auto_resize (CLUTTER_CAIRO_TEXTURE
                                                       (priv->cairo_texture), TRUE);
                g_signal_connect (priv->cairo_texture, "draw", G_CALLBACK (ckd_page_render), self);
                break;
        default:
                G_OBJECT_WARN_INVALID_PROPERTY_ID (obj, prop_id, pspec);
                break;
        }
}

static void
ckd_page_get_property (GObject *obj, guint prop_id, GValue *value, GParamSpec *pspec)
{
        CkdPage *self = CKD_PAGE (obj);
        CkdPagePriv *priv = CKD_PAGE_GET_PRIVATE (self);

        switch (prop_id) {
        case PROP_PDF_PAGE:
                g_value_set_pointer (value, priv->pdf_page);
                break;
        default:
                G_OBJECT_WARN_INVALID_PROPERTY_ID (obj, prop_id, pspec);
                break;
        }
}

static void
ckd_page_destroy (ClutterActor *self)
{
        CkdPagePriv *priv = CKD_PAGE_GET_PRIVATE (self);

        if (priv->cairo_texture) {
                clutter_actor_destroy (priv->cairo_texture);
                priv->cairo_texture = NULL;
        }

        if (CLUTTER_ACTOR_CLASS (ckd_page_parent_class)->destroy)
                CLUTTER_ACTOR_CLASS (ckd_page_parent_class)->destroy (self);
}

static void
ckd_page_get_size (PopplerPage *page, gfloat *width, gfloat *height, gfloat *aspect)
{
        gdouble w, h, a;

        poppler_page_get_size (page, &w, &h);
        if (h < G_MINFLOAT)
                h = G_MINFLOAT;
        a = w / h;

        if (width)
                *width = w;
        if (height)
                *height = h;
        if (aspect)
                *aspect = a;
}

static void
ckd_page_get_parent_size (ClutterActor *actor, gfloat *width, gfloat *height, gfloat *aspect)
{
        gfloat w, h, a;
        ClutterActor *parent = clutter_actor_get_parent (actor);
        ClutterActorBox box;

        clutter_actor_get_size (parent, &w, &h);

        if (h < G_MINFLOAT)
                h = G_MINFLOAT;
        a = w / h;

        if (width)
                *width = w;
        if (height)
                *height = h;
        if (aspect)
                *aspect = a;
}

static void
ckd_page_get_preferred_width (ClutterActor *actor,
                              gfloat for_height, gfloat *min_width, gfloat *natural_width)
{
        CkdPage *self = CKD_PAGE (actor);
        CkdPagePriv *priv = CKD_PAGE_GET_PRIVATE (self);

        gfloat w, h, a;
        ckd_page_get_parent_size (actor, &w, &h, &a);

        gfloat pw, pa;
        ckd_page_get_size (priv->pdf_page, &pw, NULL, &pa);

        if (a > pa)
                w = h * pa;

        *min_width = 0;
        *natural_width = w;

        priv->scale = w / pw;
}

static void
ckd_page_get_preferred_height (ClutterActor *actor,
                               gfloat for_width, gfloat *min_height, gfloat *natural_height)
{
        CkdPage *self = CKD_PAGE (actor);
        CkdPagePriv *priv = CKD_PAGE_GET_PRIVATE (self);

        gfloat w, h, a;
        ckd_page_get_parent_size (actor, &w, &h, &a);

        gfloat ph, pa;
        ckd_page_get_size (priv->pdf_page, NULL, &ph, &pa);

        if (a <= pa)
                h = w / pa;

        *min_height = 0;
        *natural_height = h;

        priv->scale = h / ph;
}

static void
ckd_page_allocate (ClutterActor *actor, const ClutterActorBox *box, ClutterAllocationFlags flags)
{
        CkdPage *self = CKD_PAGE (actor);
        CkdPagePriv *priv = CKD_PAGE_GET_PRIVATE (self);

        CLUTTER_ACTOR_CLASS (ckd_page_parent_class)->allocate (actor, box, flags);

        gfloat w = clutter_actor_box_get_width (box);
        gfloat h = clutter_actor_box_get_height (box);
        ClutterActorBox page_box = { 0, 0, w, h };

        clutter_actor_allocate (priv->cairo_texture, &page_box, flags);
}

static void
ckd_page_class_init (CkdPageClass *klass)
{
        g_type_class_add_private (klass, sizeof (CkdPagePriv));

        GObjectClass *base_class = G_OBJECT_CLASS (klass);
        base_class->set_property = ckd_page_set_property;
        base_class->get_property = ckd_page_get_property;

        GParamSpec *pspec;
        pspec = g_param_spec_pointer ("pdf-page",
                                      "PDF Page",
                                      "A PDF Page which to be convert to CairoTexture",
                                      G_PARAM_READABLE | G_PARAM_WRITABLE | G_PARAM_CONSTRUCT);
        g_object_class_install_property (base_class, PROP_PDF_PAGE, pspec);

        ClutterActorClass *actor_class = CLUTTER_ACTOR_CLASS (klass);
        actor_class->destroy = ckd_page_destroy;
        actor_class->paint = ckd_page_paint;
        actor_class->get_preferred_height = ckd_page_get_preferred_height;
        actor_class->get_preferred_width = ckd_page_get_preferred_width;
        actor_class->allocate = ckd_page_allocate;
}

static void
ckd_page_init (CkdPage *self)
{
        CkdPagePriv *priv = CKD_PAGE_GET_PRIVATE (self);
        priv->scale = 1.0;
}

ClutterActor *
ckd_page_new (PopplerPage *page)
{
        ClutterActor *ckd_page;
        ckd_page = g_object_new (CKD_TYPE_PAGE, "pdf-page", page, NULL);

        return ckd_page;
}
